/* PulseAudioTargetDataLine.java
   Copyright (C) 2008 Red Hat, Inc.

This file is part of IcedTea.

IcedTea is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

IcedTea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with IcedTea; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
 */

package org.classpath.icedtea.pulseaudio;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.TargetDataLine;

public class PulseAudioTargetDataLine extends PulseAudioDataLine implements
		TargetDataLine {

	private long currentFramePosition = 0;

	public PulseAudioTargetDataLine(EventLoop eventLoop, AudioFormat[] formats,
			AudioFormat defaultFormat) {
		supportedFormats = formats;
		this.eventLoop = eventLoop;
		this.defaultFormat = defaultFormat;
		this.currentFormat = defaultFormat;

	}

	@Override
	public void close() {
		PulseAudioMixer parentMixer = PulseAudioMixer.getInstance();
		parentMixer.removeTargetLine(this);

		super.close();
	}

	@Override
	public void open(AudioFormat format, int bufferSize)
			throws LineUnavailableException {
		super.open(format, bufferSize);

		PulseAudioMixer parentMixer = PulseAudioMixer.getInstance();
		parentMixer.addTargetLine(this);
	}

	protected void connectLine(int bufferSize) throws LineUnavailableException {
		StreamBufferAttributes bufferAttributes = new StreamBufferAttributes(
				bufferSize, 0, 0, 0, bufferSize / 10);
		synchronized (eventLoop.threadLock) {
			stream.connectForRecording(Stream.DEFAULT_DEVICE, bufferAttributes);
		}
	}

	@Override
	public int read(byte[] data, int offset, int length) {

		if (!isOpen) {
			throw new IllegalStateException("must call open() before read()");
		}

		if (!isStarted) {
			throw new IllegalStateException("must call start() before read()");
		}

		int frameSize = currentFormat.getFrameSize();

		if (length % frameSize != 0) {
			throw new IllegalArgumentException(
					"amount of data to read does not represent an integral number of frames");
		}

		if (length < 0) {
			throw new IllegalArgumentException("length is negative");
		}

		if (length + offset > data.length) {
			throw new ArrayIndexOutOfBoundsException("index: "
					+ (length + offset) + " array size: " + data.length);
		}

		int position = offset;
		int remainingLength = length;
		int sizeRead = 0;

		while (remainingLength != 0) {

			synchronized (eventLoop.threadLock) {
				int bytesRead = stream.read(data, remainingLength, position);
				if (bytesRead < 0) {
					return sizeRead;
				}
				sizeRead += bytesRead;
				position += bytesRead;
				remainingLength -= bytesRead;
				currentFramePosition = bytesRead / currentFormat.getFrameSize();

			}
		}

		// all the data should have been played by now
		assert (sizeRead == length);
		// currentFramePosition += (sizeWritten/getFormat().getFrameSize());
		/*
		 * FIXME when the stream is flushed() etc, instead of returning length
		 * this should unblock and return the the size of data written so far
		 */
		return sizeRead;

	}

	@Override
	public void drain() {
		// FIXME how do we drain a target data line?
	}

	@Override
	public void flush() {
		// FIXME how to flush a target data line
	}

	public int available() {
		synchronized (eventLoop.threadLock) {
			return stream.getReableSize();
		}
	}

	public int getFramePosition() {
		return (int) currentFramePosition;
	}

	public long getLongFramePosition() {
		return currentFramePosition;
	}

	public long getMicrosecondPosition() {
		return (long) (currentFramePosition / currentFormat.getFrameRate());
	}

	/*
	 * A TargetData starts when we ask it to and continues playing until we ask
	 * it to stop. there are no buffer underruns/overflows or anything so we
	 * will just fire the LineEvents manually
	 */

	@Override
	public void start() {
		super.start();

		fireLineEvent(new LineEvent(this, LineEvent.Type.START,
				AudioSystem.NOT_SPECIFIED));
	}

	@Override
	public void stop() {
		super.stop();

		fireLineEvent(new LineEvent(this, LineEvent.Type.STOP,
				AudioSystem.NOT_SPECIFIED));
	}

}